On Fri, 20 Mar 2026 at 20:20, Tomas Vondra <[email protected]> wrote:
It'd be very helpful if there was some sort of PoC
support on the pooler/client side, so that I can experiment with it and
see how helpful the new protocol message is. But I realize that's a bit
too much to ask for.
I'll see if I can whip something up, it shouldn't be too hard.
Why not to have a pg_goaway_backend() function, that'd send the
message to a single backend?
I like this idea a lot. So I added it in the attached v8 patch. This
also allowed we me to add low level tests using the libpq_pipeline
testsuite.
* In fact, does it improve the smart shutdown case in practice? Let's
say we have a single instance, and we're restarting it. It'll send
GoAway to all the clients, the good clients will try to reconnect. But
if there's even a single "bad" client ignoring the GoAway, all the
well-behaved clients will get stuck. Ofc, that can happen without the
GoAway message too - a client may disconnect because of timeout etc. But
it makes it more likely, and it'll affect the well-behaved clients.
For primary server restarts, I don't think anyone should be using smart
shutdown right now either. Any new connections to the database will be
failing for an indeterminate amount of time. I agree that sending GoAway
might worsen the problem in some cases, but it's already terrible to
start with. Fast shutdown is the only sensible restart mode for a
primary server. This seems to be generally accepted knowledge, given
that we use SIGINT (fast shutdown) in our systemd example[1].
Sending a GoAway on smart shutdown makes that shutdown mode very useful
for read replicas during a planned switch-over to another replica. Now
clients can finish their work and quickly reconnect to the new read
replica, minimizing switchover time while preventing errors.
Even when restarting primary servers, triggering a smart shutdown has a
benefits, as long as it's followed by a fast shutdown after a short
delay (e.g., 1 second). This causes slightly longer downtime (the
additional delay), but it allows most clients to disconnect on their own
terms instead of in the middle of a query. Connection errors can often
be retried transparently more easily than errors in the middle of a
query. In effect, for many applications, this could mean a reduction in
errors and only an increase in latency during a restart.
* Would it make sense to have some payload in the GoAway message? I'm
thinking about (a) some deadline by which the client should disconnect,
e.g. time of planned restart / shutdown, (b) priority, expressing how
much the client should try to disconnect (and maybe take more drastic
actions).
I thought some more about this, but ultimately, the payloads you suggest
only seem useful if a client has something inbetween "disconnect hard
now" and "disconnect when the connection is unused". I cannot think of
any such cases. i.e. what other "drastic actions" could a client take
instead of simply closing the connection. If that's the only
possibility, why not simply have the server close the connection in that
case.
Overall, I agree that having no payload in this new message feels a bit
weird. But ultimately, clients don't need any payload to do something
useful.
Also, two minor comments:
Fixed.
From aae99054edbb594fceac18c26ccbdca38de63980 Mon Sep 17 00:00:00 2001
From: Jelte Fennema-Nio <[email protected]>
Date: Sun, 19 Oct 2025 00:30:48 +0200
Subject: [PATCH v8] Add GoAway protocol message for graceful but fast server
shutdown/switchover
This commit introduces a new GoAway backend-to-frontend protocol
message (byte 'g') that the server can send to the client to politely
request that client to disconnect/reconnect when convenient. This message is
advisory only - the connection remains fully functional and clients may
continue executing queries and starting new transactions. "When
convenient" is obviously not very well defined, but the primary target
clients are clients that maintain a connection pool. Such clients should
disconnect/reconnect a connection in the pool when there's no user of
that connection. This is similar to how such clients often currently
remove a connection from the pool after the connection hits a maximum
lifetime of e.g. 1 hour.
A new pg_goaway_backend function is introduced to tell a backend to send
the GoAway message to its connected client. This makes it effectively
graceful counterpart to pg_terminate_backend. This function can be
useful to trigger a graceful reconnect. A reason this might be useful is
if it is using a lot of memory, and an operator wants the process to
clear its catalog caches.
Apart from being sent by pg_goaway_backend, the GoAway message is used by
Postgres during the already existing "smart" shutdown procedure (i.e.
when postmaster receives SIGTERM). When Postgres is in "smart" shutdown
mode existing clients can continue to run queries as usual but new
connection attempts are rejected. This mode is primarily useful when
triggering a switchover of a read replica. A load balancer can route new
connections only to the new read replica, while the old load balancer
keeps serving the existing connections until they disconnect. The
problem is that this draining of connections could often take a long
time. Even when clients only run very short queries/transactions because
the session can be kept open much longer (many connection pools use 1
hour max lifetime of a connection by default). With the introduction of
the GoAway message Postgres now sends this message to all connected
clients when it enters smart shutdown mode. If these clients respond to
the message by reconnecting/disconnecting earlier than their maximum
connection lifetime the draining can complete much quicker. Similar
benefits to switchover duration can be achieved for other applications
or proxies implementing the Postgres protocol, like when switching over
a cluster of PgBouncer machines to a newer version.
Applications/clients that use libpq can periodically check the result of
PQgoAwayReceived() at an inactive time to see whether they are asked to
reconnect.
---
doc/src/sgml/func/func-admin.sgml | 33 +++++
doc/src/sgml/libpq.sgml | 52 ++++++++
doc/src/sgml/protocol.sgml | 70 ++++++++++-
src/backend/postmaster/postmaster.c | 26 ++++
src/backend/storage/ipc/procsignal.c | 3 +
src/backend/storage/ipc/signalfuncs.c | 119 ++++++++++++++++--
src/backend/tcop/backend_startup.c | 21 +++-
src/backend/tcop/postgres.c | 36 ++++++
src/bin/psql/common.c | 7 ++
src/bin/psql/meson.build | 1 +
src/bin/psql/t/040_goaway.pl | 46 +++++++
src/include/catalog/pg_proc.dat | 5 +
src/include/libpq/libpq-be.h | 5 +
src/include/libpq/protocol.h | 1 +
src/include/storage/procsignal.h | 3 +-
src/include/tcop/tcopprot.h | 1 +
src/interfaces/libpq/exports.txt | 1 +
src/interfaces/libpq/fe-connect.c | 1 +
src/interfaces/libpq/fe-exec.c | 27 ++++
src/interfaces/libpq/fe-protocol3.c | 39 +++++-
src/interfaces/libpq/libpq-fe.h | 3 +
src/interfaces/libpq/libpq-int.h | 1 +
.../modules/libpq_pipeline/libpq_pipeline.c | 73 +++++++++++
23 files changed, 553 insertions(+), 21 deletions(-)
create mode 100644 src/bin/psql/t/040_goaway.pl
diff --git a/doc/src/sgml/func/func-admin.sgml b/doc/src/sgml/func/func-admin.sgml
index 210b1118bdf..3338169bebb 100644
--- a/doc/src/sgml/func/func-admin.sgml
+++ b/doc/src/sgml/func/func-admin.sgml
@@ -163,6 +163,39 @@
</para></entry>
</row>
+ <row>
+ <entry role="func_table_entry"><para role="func_signature">
+ <indexterm>
+ <primary>pg_goaway_backend</primary>
+ </indexterm>
+ <function>pg_goaway_backend</function> ( <parameter>pid</parameter> <type>integer</type>, <parameter>timeout</parameter> <type>bigint</type> <literal>DEFAULT</literal> <literal>0</literal> )
+ <returnvalue>boolean</returnvalue>
+ </para>
+ <para>
+ Requests the backend process with the specified process ID to send a
+ <literal>GoAway</literal> message to its client, politely asking it to
+ disconnect when convenient. The connection remains fully functional
+ after receiving the message. The same permission rules as
+ <function>pg_cancel_backend</function> apply.
+ </para>
+ <para>
+ If <parameter>timeout</parameter> is not specified or zero, this
+ function returns <literal>true</literal> whether the client actually
+ disconnects or not, indicating only that the request was successfully
+ sent to the backend. If the <parameter>timeout</parameter> is
+ specified (in milliseconds) and greater than zero, the function waits
+ until the backend has exited. If it exits within the timeout, the
+ function returns <literal>true</literal>. On timeout, a warning is
+ emitted and <literal>false</literal> is returned.
+ </para>
+ <para>
+ This is useful for gracefully draining connections from a server, for
+ example during a switchover. Clients that do not support
+ the <literal>goaway</literal> <link linkend="protocol-extensions-table">protocol extension</link> will silently ignore
+ the request.
+ </para></entry>
+ </row>
+
<row>
<entry role="func_table_entry"><para role="func_signature">
<indexterm>
diff --git a/doc/src/sgml/libpq.sgml b/doc/src/sgml/libpq.sgml
index 6db823808fc..0b408665a6f 100644
--- a/doc/src/sgml/libpq.sgml
+++ b/doc/src/sgml/libpq.sgml
@@ -2248,6 +2248,14 @@ postgresql://%2Fvar%2Flib%2Fpostgresql/dbname
equivalent to the latest protocol version supported by the libpq
version being used, which is currently <literal>3.2</literal>.
</para>
+
+ <para>
+ When using protocol version <literal>3.2</literal> or higher,
+ <application>libpq</application> requests all supported
+ <link linkend="protocol-extensions">protocol extensions</link>
+ from the server, such as <literal>goaway</literal>
+ which enables <xref linkend="libpq-PQgoAwayReceived"/>.
+ </para>
</listitem>
</varlistentry>
@@ -3004,6 +3012,50 @@ int PQserverVersion(const PGconn *conn);
</listitem>
</varlistentry>
+ <varlistentry id="libpq-PQgoAwayReceived">
+ <term><function>PQgoAwayReceived</function><indexterm><primary>PQgoAwayReceived</primary></indexterm></term>
+ <listitem>
+ <para>
+ Returns true if the server has sent a <literal>GoAway</literal> message,
+ requesting the client to disconnect when convenient.
+
+<synopsis>
+int PQgoAwayReceived(PGconn *conn);
+</synopsis>
+ </para>
+
+ <para>
+ The <literal>GoAway</literal> message is sent by the server during a
+ smart shutdown to politely request that clients disconnect. This is
+ advisory only - the connection remains fully functional and queries
+ can continue to be executed. Applications can choose to honor the request
+ by calling this function periodically and disconnect gracefully when
+ possible, such as after completing the current transaction.
+ </para>
+
+ <para>
+ This message is only sent to clients that request the
+ <literal>goaway</literal> protocol extension in the startup message,
+ which <application>libpq</application> does when using protocol version
+ 3.2 or higher (see <xref linkend="libpq-connect-max-protocol-version"/>).
+ The function returns 1 if the <literal>GoAway</literal> message was
+ received, 0 otherwise.
+ </para>
+
+ <para>
+ <function>PQgoAwayReceived</function> does not actually read data from the
+ server; it just returns messages previously absorbed by another
+ <application>libpq</application> function. So normally you would first
+ call <xref linkend="libpq-PQconsumeInput"/>, then check
+ <function>PQgoAwayReceived</function>. You can use
+ <function>select()</function> to wait for data to arrive from the
+ server, thereby using no <acronym>CPU</acronym> power unless there is
+ something to do. (See <xref linkend="libpq-PQsocket"/> to obtain the file
+ descriptor number to use with <function>select()</function>.)
+ </para>
+ </listitem>
+ </varlistentry>
+
<varlistentry id="libpq-PQerrorMessage">
<term>
<function>PQerrorMessage</function><indexterm><primary>PQerrorMessage</primary></indexterm>
diff --git a/doc/src/sgml/protocol.sgml b/doc/src/sgml/protocol.sgml
index 49f81676712..d3b3d836893 100644
--- a/doc/src/sgml/protocol.sgml
+++ b/doc/src/sgml/protocol.sgml
@@ -346,8 +346,14 @@
<tbody>
<row>
- <entry namest="last" align="center" valign="middle">
- <emphasis>(No supported protocol extensions are currently defined.)</emphasis>
+ <entry><literal>_pq_.goaway</literal></entry>
+ <entry><literal>1</literal></entry>
+ <entry>PostgreSQL 19</entry>
+ <entry>
+ Enables the server to send
+ <link linkend="protocol-message-formats-GoAway">GoAway</link>
+ messages to request the client to disconnect when convenient.
+ See <xref linkend="protocol-async"/> for more details.
</entry>
</row>
</tbody>
@@ -710,6 +716,17 @@
</para>
</listitem>
</varlistentry>
+
+ <varlistentry>
+ <term>GoAway</term>
+ <listitem>
+ <para>
+ The server requests the client politely to close the connection at its
+ earliest convenient moment. See <xref linkend="protocol-async"/> for
+ more details.
+ </para>
+ </listitem>
+ </varlistentry>
</variablelist>
</para>
@@ -1458,7 +1475,7 @@ SELCT 1/0;<!-- this typo is intentional -->
</para>
<para>
- It is possible for NoticeResponse and ParameterStatus messages to be
+ It is possible for NoticeResponse, ParameterStatus and GoAway messages to be
interspersed between CopyData messages; frontends must handle these cases,
and should be prepared for other asynchronous message types as well (see
<xref linkend="protocol-async"/>). Otherwise, any message type other than
@@ -1569,6 +1586,28 @@ SELCT 1/0;<!-- this typo is intentional -->
parameters that it does not understand or care about.
</para>
+ <para>
+ A GoAway message is sent by the server to politely request the client to
+ disconnect when convenient and reconnect when the client needs a connection
+ again. This is advisory only - the connection remains fully functional and
+ queries can continue to be executed. "When convenient" is very vaguely
+ defined on purpose because it depends on the client and application whether
+ such a moment even exists. So clients are allowed to completely ignore this
+ message and disconnect whenever they otherwise would have. An important
+ type of client that can actually honor the request to disconnect early is a
+ client that maintains a connection pool. Such a client can honor the
+ request by disconnecting a connection that has received a GoAway message
+ when it's not in use by a user of the pool. It is allowed for a server to
+ send multiple GoAway messages on the same connection, but any subsequent
+ GoAway messages after the first GoAway have no effect on the client's
+ behavior. The GoAway message is currently sent by Postgres during the
+ "smart" shutdown procedure (i.e. when postmaster receives
+ <systemitem>SIGTERM</systemitem>). The server only sends the GoAway
+ message to clients that request the <literal>goaway</literal> protocol
+ extension by setting <literal>_pq_.goaway</literal> to <literal>1</literal>
+ in the StartupMessage.
+ </para>
+
<para>
If a frontend issues a <command>LISTEN</command> command, then the
backend will send a NotificationResponse message (not to be
@@ -5359,6 +5398,31 @@ psql "dbname=postgres replication=database" -c "IDENTIFY_SYSTEM;"
</listitem>
</varlistentry>
+ <varlistentry id="protocol-message-formats-GoAway">
+ <term>GoAway (B)</term>
+ <listitem>
+ <variablelist>
+ <varlistentry>
+ <term>Byte1('g')</term>
+ <listitem>
+ <para>
+ Identifies the message as a polite request for the client to disconnect.
+ </para>
+ </listitem>
+ </varlistentry>
+
+ <varlistentry>
+ <term>Int32(4)</term>
+ <listitem>
+ <para>
+ Length of message contents in bytes, including self.
+ </para>
+ </listitem>
+ </varlistentry>
+ </variablelist>
+ </listitem>
+ </varlistentry>
+
<varlistentry id="protocol-message-formats-GSSENCRequest">
<term>GSSENCRequest (F)</term>
<listitem>
diff --git a/src/backend/postmaster/postmaster.c b/src/backend/postmaster/postmaster.c
index 3fac46c402b..b79ac383503 100644
--- a/src/backend/postmaster/postmaster.c
+++ b/src/backend/postmaster/postmaster.c
@@ -2134,7 +2134,33 @@ process_pm_shutdown_request(void)
* later state, do not change it.
*/
if (pmState == PM_RUN || pmState == PM_HOT_STANDBY)
+ {
+ dlist_iter iter;
+
connsAllowed = false;
+
+ /*
+ * Signal all backends to send a GoAway message to their
+ * clients, to politely request that they disconnect.
+ */
+ dlist_foreach(iter, &ActiveChildList)
+ {
+ PMChild *bp = dlist_container(PMChild, elem, iter.cur);
+
+ /*
+ * Only signal regular backends, since those are the ones
+ * that need to notify their clients using a GoAway
+ * message. Follow the same pattern as SignalChildren to
+ * correctly distinguish backends from WAL senders.
+ */
+ if (bp->bkend_type == B_BACKEND &&
+ !IsPostmasterChildWalSender(bp->child_slot))
+ {
+ SendProcSignal(bp->pid, PROCSIG_GOAWAY,
+ INVALID_PROC_NUMBER);
+ }
+ }
+ }
else if (pmState == PM_STARTUP || pmState == PM_RECOVERY)
{
/* There should be no clients, so proceed to stop children */
diff --git a/src/backend/storage/ipc/procsignal.c b/src/backend/storage/ipc/procsignal.c
index 7e017c8d53b..d1c9e2c44fd 100644
--- a/src/backend/storage/ipc/procsignal.c
+++ b/src/backend/storage/ipc/procsignal.c
@@ -697,6 +697,9 @@ procsignal_sigusr1_handler(SIGNAL_ARGS)
if (CheckProcSignal(PROCSIG_LOG_MEMORY_CONTEXT))
HandleLogMemoryContextInterrupt();
+ if (CheckProcSignal(PROCSIG_GOAWAY))
+ HandleGoAwayInterrupt();
+
if (CheckProcSignal(PROCSIG_PARALLEL_APPLY_MESSAGE))
HandleParallelApplyMessageInterrupt();
diff --git a/src/backend/tcop/backend_startup.c b/src/backend/tcop/backend_startup.c
index 5abf276c898..1bbd5deb448 100644
--- a/src/backend/tcop/backend_startup.c
+++ b/src/backend/tcop/backend_startup.c
@@ -779,11 +779,24 @@ ProcessStartupPacket(Port *port, bool ssl_done, bool gss_done)
{
/*
* Any option beginning with _pq_. is reserved for use as a
- * protocol-level option, but at present no such options are
- * defined.
+ * protocol-level option.
*/
- unrecognized_protocol_options =
- lappend(unrecognized_protocol_options, pstrdup(nameptr));
+ if (strcmp(nameptr, "_pq_.goaway") == 0)
+ {
+ /* Client wants to receive GoAway messages. */
+ if (strcmp(valptr, "1") != 0)
+ ereport(FATAL,
+ (errcode(ERRCODE_PROTOCOL_VIOLATION),
+ errmsg("invalid value for protocol option \"%s\": \"%s\"",
+ nameptr, valptr),
+ errhint("Valid values are: \"1\".")));
+ port->goaway_negotiated = true;
+ }
+ else
+ {
+ unrecognized_protocol_options =
+ lappend(unrecognized_protocol_options, pstrdup(nameptr));
+ }
}
else
{
diff --git a/src/backend/tcop/postgres.c b/src/backend/tcop/postgres.c
index b3563113219..c1df84cc547 100644
--- a/src/backend/tcop/postgres.c
+++ b/src/backend/tcop/postgres.c
@@ -42,8 +42,10 @@
#include "common/pg_prng.h"
#include "jit/jit.h"
#include "libpq/libpq.h"
+#include "libpq/pqcomm.h"
#include "libpq/pqformat.h"
#include "libpq/pqsignal.h"
+#include "libpq/protocol.h"
#include "mb/pg_wchar.h"
#include "mb/stringinfo_mb.h"
#include "miscadmin.h"
@@ -93,6 +95,14 @@ const char *debug_query_string; /* client-supplied query string */
/* Note: whereToSendOutput is initialized for the bootstrap/standalone case */
CommandDest whereToSendOutput = DestDebug;
+/*
+ * Track whether we've been asked to send GoAway and whether we've sent it.
+ * GoAwayPending is set by the PROCSIG_GOAWAY signal handler.
+ * GoAwaySent tracks whether we've already sent the GoAway message.
+ */
+static volatile sig_atomic_t GoAwayPending = false;
+static bool GoAwaySent = false;
+
/* flag for logging end of session */
bool Log_disconnections = false;
@@ -508,6 +518,20 @@ ProcessClientReadInterrupt(bool blocked)
/* Check for general interrupts that arrived before/while reading */
CHECK_FOR_INTERRUPTS();
+ /* Send GoAway message if requested */
+ if (GoAwayPending && !GoAwaySent &&
+ whereToSendOutput == DestRemote &&
+ MyProcPort && MyProcPort->goaway_negotiated)
+ {
+ StringInfoData buf;
+
+ pq_beginmessage(&buf, PqMsg_GoAway);
+ pq_endmessage(&buf);
+ pq_flush();
+
+ GoAwaySent = true;
+ }
+
/* Process sinval catchup interrupts, if any */
if (catchupInterruptPending)
ProcessCatchupInterrupt();
@@ -3067,6 +3091,18 @@ FloatExceptionHandler(SIGNAL_ARGS)
"invalid operation, such as division by zero.")));
}
+/*
+ * Tell the next ProcessClientReadInterrupt() call to send a GoAway message to
+ * the client. Runs in a SIGUSR1 handler.
+ */
+void
+HandleGoAwayInterrupt(void)
+{
+ GoAwayPending = true;
+ InterruptPending = true;
+ /* latch will be set by procsignal_sigusr1_handler */
+}
+
/*
* Tell the next CHECK_FOR_INTERRUPTS() to process recovery conflicts. Runs
* in a SIGUSR1 handler.
diff --git a/src/bin/psql/common.c b/src/bin/psql/common.c
index 2eadd391a9c..4e99ed69841 100644
--- a/src/bin/psql/common.c
+++ b/src/bin/psql/common.c
@@ -739,6 +739,7 @@ PSQLexecWatch(const char *query, const printQueryOpt *opt, FILE *printQueryFout,
static void
PrintNotifications(void)
{
+ static bool goaway_reported = false;
PGnotify *notify;
PQconsumeInput(pset.db);
@@ -755,6 +756,12 @@ PrintNotifications(void)
PQfreemem(notify);
PQconsumeInput(pset.db);
}
+
+ if (!goaway_reported && PQgoAwayReceived(pset.db))
+ {
+ pg_log_info("Server sent GoAway, requesting disconnect when convenient.");
+ goaway_reported = true;
+ }
}
diff --git a/src/bin/psql/meson.build b/src/bin/psql/meson.build
index 922b2845267..047b40cc40e 100644
--- a/src/bin/psql/meson.build
+++ b/src/bin/psql/meson.build
@@ -78,6 +78,7 @@ tests += {
't/010_tab_completion.pl',
't/020_cancel.pl',
't/030_pager.pl',
+ 't/040_goaway.pl',
],
},
}
diff --git a/src/bin/psql/t/040_goaway.pl b/src/bin/psql/t/040_goaway.pl
new file mode 100644
index 00000000000..a53653308bf
--- /dev/null
+++ b/src/bin/psql/t/040_goaway.pl
@@ -0,0 +1,46 @@
+
+# Copyright (c) 2021-2026, PostgreSQL Global Development Group
+
+use strict;
+use warnings FATAL => 'all';
+
+use PostgreSQL::Test::Cluster;
+use PostgreSQL::Test::Utils;
+use Test::More;
+use Time::HiRes qw(usleep);
+
+my $node = PostgreSQL::Test::Cluster->new('main');
+$node->init;
+$node->start;
+
+my $psql = $node->background_psql('postgres');
+
+# Confirm connection works
+my $result = $psql->query_safe("SELECT 'before_goaway'");
+like($result, qr/before_goaway/, 'connection works before smart shutdown');
+
+# Initiate smart shutdown without waiting for it to complete
+$node->command_ok(
+ [ 'pg_ctl', 'stop', '-D', $node->data_dir, '-m', 'smart', '--no-wait' ],
+ 'pg_ctl smart shutdown');
+
+# The backend sends GoAway once it processes the smart shutdown signal.
+# Poll with queries until psql reports it.
+my $saw_goaway = 0;
+for (my $i = 0; $i < 100; $i++)
+{
+ my $out = $psql->query("SELECT 'after_goaway'");
+ if ($psql->{stderr} =~ /Server sent GoAway, requesting disconnect when convenient/)
+ {
+ $saw_goaway = 1;
+ # The query should still have succeeded
+ like($out, qr/after_goaway/, 'query still works after GoAway');
+ last;
+ }
+ usleep(50_000);
+}
+ok($saw_goaway, 'psql reported GoAway notice during smart shutdown');
+
+$psql->quit;
+
+done_testing();
diff --git a/src/include/catalog/pg_proc.dat b/src/include/catalog/pg_proc.dat
index 84e7adde0e5..e8584a3e458 100644
--- a/src/include/catalog/pg_proc.dat
+++ b/src/include/catalog/pg_proc.dat
@@ -6765,6 +6765,11 @@
proargtypes => 'int4 int8', proargnames => '{pid,timeout}',
proargdefaults => '{0}',
prosrc => 'pg_terminate_backend' },
+{ oid => '9018', descr => 'request a server process to send GoAway to its client',
+ proname => 'pg_goaway_backend', provolatile => 'v', prorettype => 'bool',
+ proargtypes => 'int4 int8', proargnames => '{pid,timeout}',
+ proargdefaults => '{0}',
+ prosrc => 'pg_goaway_backend' },
{ oid => '2172', descr => 'prepare for taking an online backup',
proname => 'pg_backup_start', provolatile => 'v', proparallel => 'r',
prorettype => 'pg_lsn', proargtypes => 'text bool',
diff --git a/src/include/libpq/libpq-be.h b/src/include/libpq/libpq-be.h
index 921b2daa4ff..b8769412c72 100644
--- a/src/include/libpq/libpq-be.h
+++ b/src/include/libpq/libpq-be.h
@@ -202,6 +202,11 @@ typedef struct Port
void *gss;
#endif
+ /*
+ * Protocol extensions.
+ */
+ bool goaway_negotiated; /* client supports GoAway message */
+
/*
* SSL structures.
*/
diff --git a/src/include/libpq/protocol.h b/src/include/libpq/protocol.h
index eae8f0e7238..fd3c195daa2 100644
--- a/src/include/libpq/protocol.h
+++ b/src/include/libpq/protocol.h
@@ -53,6 +53,7 @@
#define PqMsg_FunctionCallResponse 'V'
#define PqMsg_CopyBothResponse 'W'
#define PqMsg_ReadyForQuery 'Z'
+#define PqMsg_GoAway 'g'
#define PqMsg_NoData 'n'
#define PqMsg_PortalSuspended 's'
#define PqMsg_ParameterDescription 't'
diff --git a/src/include/storage/procsignal.h b/src/include/storage/procsignal.h
index 348fba53a93..ecbc62b7fd7 100644
--- a/src/include/storage/procsignal.h
+++ b/src/include/storage/procsignal.h
@@ -39,9 +39,10 @@ typedef enum
PROCSIG_RECOVERY_CONFLICT, /* backend is blocking recovery, check
* PGPROC->pendingRecoveryConflicts for the
* reason */
+ PROCSIG_GOAWAY, /* request backend to send GoAway to client */
} ProcSignalReason;
-#define NUM_PROCSIGNALS (PROCSIG_RECOVERY_CONFLICT + 1)
+#define NUM_PROCSIGNALS (PROCSIG_GOAWAY + 1)
typedef enum
{
diff --git a/src/include/tcop/tcopprot.h b/src/include/tcop/tcopprot.h
index 5bc5bcfb20d..a36547a11f7 100644
--- a/src/include/tcop/tcopprot.h
+++ b/src/include/tcop/tcopprot.h
@@ -74,6 +74,7 @@ extern void die(SIGNAL_ARGS);
pg_noreturn extern void quickdie(SIGNAL_ARGS);
extern void StatementCancelHandler(SIGNAL_ARGS);
pg_noreturn extern void FloatExceptionHandler(SIGNAL_ARGS);
+extern void HandleGoAwayInterrupt(void);
extern void HandleRecoveryConflictInterrupt(void);
extern void ProcessClientReadInterrupt(bool blocked);
extern void ProcessClientWriteInterrupt(bool blocked);
diff --git a/src/interfaces/libpq/exports.txt b/src/interfaces/libpq/exports.txt
index 1e3d5bd5867..f57aae46b65 100644
--- a/src/interfaces/libpq/exports.txt
+++ b/src/interfaces/libpq/exports.txt
@@ -211,3 +211,4 @@ PQdefaultAuthDataHook 208
PQfullProtocolVersion 209
appendPQExpBufferVA 210
PQgetThreadLock 211
+PQgoAwayReceived 212
diff --git a/src/interfaces/libpq/fe-connect.c b/src/interfaces/libpq/fe-connect.c
index db9b4c8edbf..4d81d5f03bb 100644
--- a/src/interfaces/libpq/fe-connect.c
+++ b/src/interfaces/libpq/fe-connect.c
@@ -701,6 +701,7 @@ pqDropServerData(PGconn *conn)
conn->password_needed = false;
conn->gssapi_used = false;
conn->write_failed = false;
+ conn->goaway_received = false;
free(conn->write_err_msg);
conn->write_err_msg = NULL;
conn->oauth_want_retry = false;
diff --git a/src/interfaces/libpq/fe-exec.c b/src/interfaces/libpq/fe-exec.c
index 203d388bdbf..bc1d95324ed 100644
--- a/src/interfaces/libpq/fe-exec.c
+++ b/src/interfaces/libpq/fe-exec.c
@@ -2702,6 +2702,33 @@ PQnotifies(PGconn *conn)
return event;
}
+/*
+ * PQgoAwayReceived
+ * returns 1 if a GoAway message has been received from the server
+ * returns 0 if not
+ *
+ * Note that this function does not read any new data from the socket;
+ * caller should call PQconsumeInput() first if they want to ensure
+ * all available data has been read.
+ */
+int
+PQgoAwayReceived(PGconn *conn)
+{
+ if (!conn)
+ return 0;
+
+ if (conn->goaway_received)
+ return 1;
+
+ /*
+ * Parse any available data to see if a GoAway message has arrived.
+ */
+ parseInput(conn);
+
+ return conn->goaway_received ? 1 : 0;
+}
+
+
/*
* PQputCopyData - send some data to the backend during COPY IN or COPY BOTH
*
diff --git a/src/interfaces/libpq/fe-protocol3.c b/src/interfaces/libpq/fe-protocol3.c
index 8c1fda5caf0..d09002a22ac 100644
--- a/src/interfaces/libpq/fe-protocol3.c
+++ b/src/interfaces/libpq/fe-protocol3.c
@@ -134,8 +134,8 @@ pqParseInput3(PGconn *conn)
}
/*
- * NOTIFY and NOTICE messages can happen in any state; always process
- * them right away.
+ * NOTIFY and NOTICE and GoAway messages can happen in any state;
+ * always process them right away.
*
* Most other messages should only be processed while in BUSY state.
* (In particular, in READY state we hold off further parsing until
@@ -159,6 +159,10 @@ pqParseInput3(PGconn *conn)
if (pqGetErrorNotice3(conn, false))
return;
}
+ else if (id == PqMsg_GoAway)
+ {
+ conn->goaway_received = true;
+ }
else if (conn->asyncStatus != PGASYNC_BUSY)
{
/* If not IDLE state, just wait ... */
@@ -1446,6 +1450,7 @@ pqGetNegotiateProtocolVersion3(PGconn *conn)
int num;
bool found_test_protocol_negotiation;
bool expect_test_protocol_negotiation;
+ bool requested_goaway = conn->pversion >= PG_PROTOCOL(3, 2);
/*
* During 19beta only, if protocol grease is in use, assume that it's the
@@ -1521,8 +1526,7 @@ pqGetNegotiateProtocolVersion3(PGconn *conn)
conn->pversion = their_version;
/*
- * Check that all expected unsupported parameters are reported by the
- * server.
+ * Process the list of unsupported protocol extensions.
*/
found_test_protocol_negotiation = false;
expect_test_protocol_negotiation = (conn->max_pversion == PG_PROTOCOL_GREASE);
@@ -1539,12 +1543,23 @@ pqGetNegotiateProtocolVersion3(PGconn *conn)
goto failure;
}
- /* Check if this is the expected test parameter */
+ /*
+ * Handle protocol extensions that we requested. Extensions that were
+ * not requested by us are an error.
+ */
if (expect_test_protocol_negotiation &&
strcmp(conn->workBuffer.data, "_pq_.test_protocol_negotiation") == 0)
{
found_test_protocol_negotiation = true;
}
+ else if (requested_goaway && strcmp(conn->workBuffer.data, "_pq_.goaway") == 0)
+ {
+ /*
+ * Server doesn't support GoAway, that's fine. It will simply never
+ * send the message.
+ */
+ continue;
+ }
else
{
libpq_append_conn_error(conn, "received invalid protocol negotiation message: server reported an unsupported parameter that was not requested (\"%s\")",
@@ -1905,6 +1920,9 @@ getCopyDataMessage(PGconn *conn)
if (getParameterStatus(conn))
return 0;
break;
+ case PqMsg_GoAway:
+ conn->goaway_received = true;
+ break;
case PqMsg_CopyData:
return msgLength;
case PqMsg_CopyDone:
@@ -2398,6 +2416,9 @@ pqFunctionCall3(PGconn *conn, Oid fnid,
if (getParameterStatus(conn))
continue;
break;
+ case PqMsg_GoAway:
+ conn->goaway_received = true;
+ break;
default:
/* The backend violates the protocol. */
libpq_append_conn_error(conn, "protocol error: id=0x%x", id);
@@ -2531,6 +2552,14 @@ build_startup_packet(const PGconn *conn, char *packet,
}
}
+ /*
+ * Request protocol extensions. Only do this for protocol version 3.2 and
+ * later, to avoid confusing old proxies that don't understand _pq_.*
+ * options.
+ */
+ if (conn->pversion >= PG_PROTOCOL(3, 2))
+ ADD_STARTUP_OPTION("_pq_.goaway", "1");
+
/* Add trailing terminator */
if (packet)
packet[packet_len] = '\0';
diff --git a/src/interfaces/libpq/libpq-fe.h b/src/interfaces/libpq/libpq-fe.h
index f06e7a972c3..d1a6dd54407 100644
--- a/src/interfaces/libpq/libpq-fe.h
+++ b/src/interfaces/libpq/libpq-fe.h
@@ -68,6 +68,8 @@ extern "C"
#define LIBPQ_HAS_GET_THREAD_LOCK 1
/* Indicates presence of the PQAUTHDATA_OAUTH_BEARER_TOKEN_V2 authdata hook */
#define LIBPQ_HAS_OAUTH_BEARER_TOKEN_V2 1
+/* Indicates presence of PQgoAwayReceived */
+#define LIBPQ_HAS_GOAWAY 1
/*
* Option flags for PQcopyResult
@@ -419,6 +421,7 @@ extern const char *PQparameterStatus(const PGconn *conn,
extern int PQprotocolVersion(const PGconn *conn);
extern int PQfullProtocolVersion(const PGconn *conn);
extern int PQserverVersion(const PGconn *conn);
+extern int PQgoAwayReceived(PGconn *conn);
extern char *PQerrorMessage(const PGconn *conn);
extern int PQsocket(const PGconn *conn);
extern int PQbackendPID(const PGconn *conn);
diff --git a/src/interfaces/libpq/libpq-int.h b/src/interfaces/libpq/libpq-int.h
index bd7eb59f5f8..2018927d0cb 100644
--- a/src/interfaces/libpq/libpq-int.h
+++ b/src/interfaces/libpq/libpq-int.h
@@ -511,6 +511,7 @@ struct pg_conn
bool sigpipe_flag; /* can we mask SIGPIPE via MSG_NOSIGNAL? */
bool write_failed; /* have we had a write failure on sock? */
char *write_err_msg; /* write error message, or NULL if OOM */
+ bool goaway_received; /* true if server sent GoAway message */
bool auth_required; /* require an authentication challenge from
* the server? */
diff --git a/src/test/modules/libpq_pipeline/libpq_pipeline.c b/src/test/modules/libpq_pipeline/libpq_pipeline.c
index aa0a6bbe762..f2b58d53392 100644
--- a/src/test/modules/libpq_pipeline/libpq_pipeline.c
+++ b/src/test/modules/libpq_pipeline/libpq_pipeline.c
@@ -2101,6 +2101,76 @@ process_result(PGconn *conn, PGresult *res, int results, int numsent)
}
+static void
+test_goaway(PGconn *conn)
+{
+ PGconn *otherConn;
+ PGresult *res;
+ int pid;
+ char pid_str[32];
+ int i;
+
+ fprintf(stderr, "test goaway... ");
+
+ otherConn = copy_connection(conn);
+ Assert(PQstatus(otherConn) == CONNECTION_OK);
+
+ if (!PQconsumeInput(conn))
+ pg_fatal("PQconsumeInput failed: %s", PQerrorMessage(conn));
+
+ pid = PQbackendPID(conn);
+ snprintf(pid_str, sizeof(pid_str), "%d", pid);
+
+ /* Verify GoAway has not been received yet */
+ if (PQgoAwayReceived(conn))
+ pg_fatal("GoAway received before sending request");
+
+ /* Ask the target backend to send GoAway */
+ {
+ const char *paramValues[1] = {pid_str};
+
+ res = PQexecParams(otherConn,
+ "SELECT pg_goaway_backend($1)",
+ 1, NULL, paramValues,
+ NULL, NULL, 0);
+ }
+ if (PQresultStatus(res) != PGRES_TUPLES_OK)
+ pg_fatal("pg_goaway_backend failed: %s", PQerrorMessage(otherConn));
+ if (strcmp(PQgetvalue(res, 0, 0), "t") != 0)
+ pg_fatal("pg_goaway_backend returned false");
+ PQclear(res);
+
+ /*
+ * The GoAway signal is delivered asynchronously. Poll PQgoAwayReceived on
+ * the target connection until it returns true.
+ */
+ for (i = 0; i < 100; i++)
+ {
+ if (!PQconsumeInput(conn))
+ pg_fatal("PQconsumeInput failed: %s", PQerrorMessage(conn));
+
+ if (PQgoAwayReceived(conn))
+ break;
+
+ pg_usleep(50000); /* 50ms */
+ }
+
+ if (!PQgoAwayReceived(conn))
+ pg_fatal("GoAway not received after sending pg_goaway_backend");
+
+ /* Verify the connection still works after GoAway */
+ res = PQexec(conn, "SELECT 'still_alive'");
+ if (PQresultStatus(res) != PGRES_TUPLES_OK)
+ pg_fatal("query failed after GoAway: %s", PQerrorMessage(conn));
+ if (strcmp(PQgetvalue(res, 0, 0), "still_alive") != 0)
+ pg_fatal("unexpected query result after GoAway");
+ PQclear(res);
+
+ PQfinish(otherConn);
+
+ fprintf(stderr, "ok\n");
+}
+
static void
usage(const char *progname)
{
@@ -2118,6 +2188,7 @@ print_test_list(void)
{
printf("cancel\n");
printf("disallowed_in_pipeline\n");
+ printf("goaway\n");
printf("multi_pipelines\n");
printf("nosync\n");
printf("pipeline_abort\n");
@@ -2225,6 +2296,8 @@ main(int argc, char **argv)
test_cancel(conn);
else if (strcmp(testname, "disallowed_in_pipeline") == 0)
test_disallowed_in_pipeline(conn);
+ else if (strcmp(testname, "goaway") == 0)
+ test_goaway(conn);
else if (strcmp(testname, "multi_pipelines") == 0)
test_multi_pipelines(conn);
else if (strcmp(testname, "nosync") == 0)
base-commit: 322bab79744dfb8f7ddb5191b3102cf7986d14a0
diff --git a/src/backend/storage/ipc/signalfuncs.c b/src/backend/storage/ipc/signalfuncs.c
index 800b699de21..4bd31c36a40 100644
--- a/src/backend/storage/ipc/signalfuncs.c
+++ b/src/backend/storage/ipc/signalfuncs.c
@@ -23,15 +23,17 @@
#include "storage/pmsignal.h"
#include "storage/proc.h"
#include "storage/procarray.h"
+#include "storage/procsignal.h"
#include "utils/acl.h"
#include "utils/fmgrprotos.h"
#include "utils/wait_event.h"
/*
- * Send a signal to another backend.
+ * Check whether the current user has permission to signal the backend with
+ * the given PID.
*
- * The signal is delivered if the user is either a superuser or the same
+ * Sending a signal is allowed if the user is either a superuser or the same
* role as the backend being signaled. For "dangerous" signals, an explicit
* check for superuser needs to be done prior to calling this function.
*
@@ -42,6 +44,10 @@
* In the event of a general failure (return code 1), a warning message will
* be emitted. For permission errors, doing that is the responsibility of
* the caller.
+ *
+ * The caller can optionally obtain a pointer to the backend's PGPROC struct
+ * using the proc_return argument. If the caller doesn't need that, it can
+ * pass NULL.
*/
#define SIGNAL_BACKEND_SUCCESS 0
#define SIGNAL_BACKEND_ERROR 1
@@ -49,17 +55,22 @@
#define SIGNAL_BACKEND_NOSUPERUSER 3
#define SIGNAL_BACKEND_NOAUTOVAC 4
static int
-pg_signal_backend(int pid, int sig)
+pg_check_signal_backend(int pid, PGPROC **proc_return)
{
+
PGPROC *proc = BackendPidGetProc(pid);
+ if (proc_return != NULL)
+ *proc_return = proc;
+
/*
* BackendPidGetProc returns NULL if the pid isn't valid; but by the time
* we reach kill(), a process for which we get a valid proc here might
* have terminated on its own. There's no way to acquire a lock on an
* arbitrary process to prevent that. But since so far all the callers of
* this mechanism involve some request for ending the process anyway, that
* it might end on its own first is not a problem.
*
* Note that proc will also be NULL if the pid refers to an auxiliary
* process or the postmaster (neither of which can be signaled via
@@ -100,6 +111,22 @@ pg_signal_backend(int pid, int sig)
!has_privs_of_role(GetUserId(), ROLE_PG_SIGNAL_BACKEND))
return SIGNAL_BACKEND_NOPERMISSION;
+ return SIGNAL_BACKEND_SUCCESS;
+}
+
+/*
+ * Send a signal to another backend, after checking permissions.
+ *
+ * See pg_check_signal_backend for return codes.
+ */
+static int
+pg_signal_backend(int pid, int sig)
+{
+ int r = pg_check_signal_backend(pid, NULL);
+
+ if (r != SIGNAL_BACKEND_SUCCESS)
+ return r;
+
/*
* Can the process we just validated above end, followed by the pid being
* recycled for a new process, before reaching here? Then we'd be trying
@@ -276,6 +303,82 @@ pg_terminate_backend(PG_FUNCTION_ARGS)
PG_RETURN_BOOL(r == SIGNAL_BACKEND_SUCCESS);
}
+/*
+ * Request a backend to send a GoAway message to its client, politely asking
+ * it to disconnect when convenient. The connection remains fully functional
+ * after receiving GoAway.
+ *
+ * If timeout is 0, returns true as soon as the signal is sent successfully.
+ * If timeout is nonzero, waits up to that many milliseconds for the backend
+ * to exit (i.e. for the client to honor the GoAway and disconnect). Returns
+ * true if the backend exits within the timeout, false (with a warning) if not.
+ *
+ * Permission checking follows the same rules as pg_cancel_backend.
+ */
+Datum
+pg_goaway_backend(PG_FUNCTION_ARGS)
+{
+ int pid;
+ int r;
+ int timeout; /* milliseconds */
+ PGPROC *proc;
+
+ pid = PG_GETARG_INT32(0);
+ timeout = PG_GETARG_INT64(1);
+
+ if (timeout < 0)
+ ereport(ERROR,
+ (errcode(ERRCODE_NUMERIC_VALUE_OUT_OF_RANGE),
+ errmsg("\"timeout\" must not be negative")));
+
+ r = pg_check_signal_backend(pid, &proc);
+
+ if (r == SIGNAL_BACKEND_NOSUPERUSER)
+ ereport(ERROR,
+ (errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+ errmsg("permission denied to send GoAway"),
+ errdetail("Only roles with the %s attribute may send GoAway to backends of roles with the %s attribute.",
+ "SUPERUSER", "SUPERUSER")));
+
+ if (r == SIGNAL_BACKEND_NOAUTOVAC)
+ ereport(ERROR,
+ (errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+ errmsg("permission denied to send GoAway"),
+ errdetail("Only roles with privileges of the \"%s\" role may send GoAway to autovacuum workers.",
+ "pg_signal_autovacuum_worker")));
+
+ if (r == SIGNAL_BACKEND_NOPERMISSION)
+ ereport(ERROR,
+ (errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+ errmsg("permission denied to send GoAway"),
+ errdetail("Only roles with privileges of the role whose backend is being signaled or with privileges of the \"%s\" role may send GoAway to this backend.",
+ "pg_signal_backend")));
+
+ if (r != SIGNAL_BACKEND_SUCCESS)
+ PG_RETURN_BOOL(false);
+
+ if (proc->backendType != B_BACKEND)
+ {
+ ereport(WARNING,
+ (errmsg("PID %d is not a regular backend process", pid)));
+ PG_RETURN_BOOL(false);
+ }
+
+ if (SendProcSignal(pid, PROCSIG_GOAWAY,
+ INVALID_PROC_NUMBER) != 0)
+ {
+ ereport(WARNING,
+ (errmsg("could not send GoAway signal to process %d: %m", pid)));
+ PG_RETURN_BOOL(false);
+ }
+
+ /* Wait only if actually requested */
+ if (timeout > 0)
+ PG_RETURN_BOOL(pg_wait_until_termination(pid, timeout));
+
+ PG_RETURN_BOOL(true);
+}
+
/*
* Signal to reload the database configuration
*
@@ -49,17 +55,22 @@
/*
* BackendPidGetProc returns NULL if the pid isn't valid; but by the time
- * we reach kill(), a process for which we get a valid proc here might
- * have terminated on its own. There's no way to acquire a lock on an
- * arbitrary process to prevent that. But since so far all the callers of
- * this mechanism involve some request for ending the process anyway, that
- * it might end on its own first is not a problem.
+ * the caller actually tries to send a signal, a process for which we get
+ * a valid proc here might have terminated on its own. There's no way to
+ * acquire a lock on an arbitrary process to prevent that. But since so
+ * far all the callers of this mechanism involve some request for ending
+ * the process anyway, that it might end on its own first is not a
+ * problem.
*
* Note that proc will also be NULL if the pid refers to an auxiliary
--
2.53.0